package core;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.Locale;
import java.util.Map;
import java.util.Scanner;
import java.util.TreeMap;

import javax.swing.JOptionPane;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import core.Keypoint.ZOrder;

@SuppressWarnings({ "unchecked" })
public class IOMgr {
	private String baseDir;
	private String infoDir;	// directory containing the keypoints for all images
	private String imgDir;
	private String pbDir;
	private String allEntriesFileName;
	private ArrayList<String> entries;
	private int curEntry;
	private boolean readOnly;
	private Map<Integer, KeypointState> currentKeypoints = new TreeMap<Integer,KeypointState>();
	private String[] regionNames;
	private ArrayList<Integer>[] currentLabels;
	private Double[] currentBounds=null;
	private String currentPose=null;
	private Integer currentVOCid=null;	
//	private static final String HIDDEN_STR = "hidden";
//	private static final String VISIBLE_STR = "visible";
//	private static final String BODY_TYPE = "body_type";
	private static final String BODY_TYPE_UNKNOWN = "Unknown";
	private ConfigMgr configMgr;

	int bodyType = -1;
	public static class KeypointState {
		public KeypointState() {}
		public KeypointState(double x, double y, double z, ZOrder zorder, boolean hidden) {
			this.x = x;
			this.y = y;
			this.z = z;
			this.zorder = zorder;
			this.hidden = hidden;
		}

		double x,y,z;
		ZOrder zorder;
		boolean hidden;
	}

	public int numEntries() { return entries.size(); }
	public boolean isReadOnly() { return readOnly; }
	
	
	public String getCurImageFullName() {
		return baseDir + imgDir+ getCurImageName(true);
	}
	
	public String getCurPbImageFullName() {
		return baseDir + pbDir+ getCurImageName(false)+".png";
	}
			
	public KeypointState getKeypoint(int label) {
		return currentKeypoints.get(label);
	}
	
	public void clearKeypoints() { currentKeypoints.clear(); }
	public void setKeypoint(int label, KeypointState kp) {
		currentKeypoints.put(label, kp);
	}
	public void setBodyType(int bodyType) {
		this.bodyType = bodyType;
	}
	public int getBodyType() {
		return bodyType;
	}
	public int curEntry() { return curEntry; }
	
	public boolean nextEntry() {
		if (curEntry+1<entries.size()) {
			curEntry++;
			return syncFromDisk();
		}
		return false;
	}
	
	public boolean prevEntry() {
		if (curEntry>0) {
			curEntry--;
			return syncFromDisk();
		}
		return false;
	}
	
	public boolean gotoEntry(int entry) {
		if (entry>=0 && entry<entries.size()) {
			curEntry=entry;
			return syncFromDisk();
		}
		return false;
	}
	
	// Inserts a new keypoint set for the current image and selects it
	public boolean writeKeypointsInANewSet() {
		if (readOnly) {
			JOptionPane.showMessageDialog(null,"The demo files are read-only","I/O Error",JOptionPane.ERROR_MESSAGE);
			return false;
		}
		int oldEntry=curEntry;
		// Find an empty index
		String crImgName = getCurImageName(false)+'_';
		int i=0;
		while (entries.contains(crImgName+i+".xml")) {
			i++;
		}
		
		if (i>0 || (new File(baseDir+infoDir+entries.get(curEntry)+".xml").exists())) {
			curEntry=entries.indexOf(crImgName+(i-1)+".xml")+1;
		} else
			i=0; // if we just have an image and no annotation, place the annotation at the 0 index
		
		entries.add(curEntry,crImgName+i+".xml");
		if (!writeKeypoints()) {
			entries.remove(curEntry);
			curEntry=oldEntry;
			return false;
		}
		if (!writeAllEntries()) {
			entries.remove(curEntry);					
			boolean success=(new File((baseDir+infoDir+entries.get(curEntry)+".xml"))).delete();
			assert(success);	// if it was created successfully it must be deleted successfully
			curEntry=oldEntry;
			return false;
		}
		return true;
	}
	
	// Deletes the current keypoint set and moves to the next one or returns false if only one entry is left
	public boolean deleteCurrentKeypointSet() {
		if (readOnly) {
			JOptionPane.showMessageDialog(null,"The demo files are read-only","I/O Error",JOptionPane.ERROR_MESSAGE);
			return false;
		}
		if (entries.size()<2) {
			JOptionPane.showMessageDialog(null,"Cannot delete the last entry","I/O Error",JOptionPane.ERROR_MESSAGE);
			return false;
		}

		File f = new File(baseDir+infoDir+entries.get(curEntry));
		if (!f.isFile()) {
			JOptionPane.showMessageDialog(null,"There are no entries to delete in this image","I/O Error",JOptionPane.ERROR_MESSAGE);
			return false;
		}

		// Remove the entry, but if it is the last one for the given image, keep an entry with the given image.
		String oldElement = entries.get(curEntry);
		String thisImageName = getCurImageName(false);
		entries.remove(curEntry);
		boolean thisWasLastEntryOfImage = !entriesContainImageName(thisImageName);
		if (thisWasLastEntryOfImage)
			entries.add(curEntry,thisImageName+"_0.xml");
		
		if (!writeAllEntries()) {
			// rollback
			if (thisWasLastEntryOfImage) 
				entries.remove(curEntry);
			entries.add(curEntry,oldElement);
			return false;
		}

		if (!f.delete()) {
			// rollback
			if (thisWasLastEntryOfImage) 
				entries.remove(curEntry);
			entries.add(curEntry,oldElement);
			boolean success = writeAllEntries();
			assert(success);
			JOptionPane.showMessageDialog(null,"Could not delete file "+f.getAbsolutePath(),"I/O Error",JOptionPane.ERROR_MESSAGE);
			return false;
		}
		
		if (curEntry==entries.size())
			curEntry--;
		return syncFromDisk();
	}
	
	private boolean syncFromDisk() {
			return readCurrentAnnotation();
//		bodyType = -1;
//		for (int i=0; i<currentLabels.length; ++i)
//			currentLabels[i] = new ArrayList<Integer>();
//
//		BufferedReader reader;
//		String path = new String(baseDir+infoDir+entries.get(curEntry)+".xml");
//		try {
//			if (readOnly) {
//				if (Thread.currentThread().getContextClassLoader().getResource(path)==null) {
//					currentKeypoints.clear();
//					return true;					
//				}
//				reader = new BufferedReader(new InputStreamReader(Thread.currentThread().getContextClassLoader().getResourceAsStream(path)));
//			} else {
//				File f = new File(path);
//				if (!f.exists()) {
//					if (!new File(baseDir+infoDir).exists()) {
//						(new File(baseDir+infoDir)).mkdir();
//					}
//					currentKeypoints.clear();
//					return true;					
//				}
//				reader = new BufferedReader(new FileReader(f));
//			}
//		} catch (Exception e) {
//			JOptionPane.showMessageDialog(null,"Cannot read "+path,"I/O Error",JOptionPane.ERROR_MESSAGE);
//			return false;
//		}
//		
//		currentKeypoints.clear();
//		
//		try {
//			String line=null;
//			while ((line=reader.readLine())!=null) {
////				System.out.println("Read: "+line);
//				Scanner s = new Scanner(line);
//				s.useLocale(Locale.US);
//				String name = s.next();
////				System.out.println("name: "+name);
//				if (name.equals(BODY_TYPE)) {
//					name = s.next();
//					if (!name.equals(BODY_TYPE_UNKNOWN)) {
//						for (int i=0; i<configMgr.getSubcategoryNames().length; ++i)
//							if (name.equals(configMgr.getSubcategoryNames()[i]))
//								bodyType=i;
//						if (bodyType==-1) {
//							JOptionPane.showMessageDialog(null,"Cannot read body type from "+infoDir+entries.get(curEntry)+".xml","I/O Error",JOptionPane.ERROR_MESSAGE);
//						}
//					}
//					continue;
//				}
//
//				int keypointVal = configMgr.getKeypointID(name);
//				if (keypointVal!=-1) {
//					currentKeypoints.put(keypointVal, readKeypointState(s));
//					continue;
//				}
//				
//				int regionID = getRegionNameID(name);
//				if (regionID>=0) {
////					System.out.println("RegionID="+regionID);
//					while (s.hasNextInt()) {
//						currentLabels[regionID].add(s.nextInt());
//					}
//	//				System.out.println("RegionID="+regionID+" labels:"+currentLabels[regionID]);
//				} else {
//					System.out.println("Warning: Ignoring unrecognized region name: "+name);
//				}
//			}
//			reader.close();
//		} catch (Exception e) {
//			System.out.println("Exception: "+e);
//			JOptionPane.showMessageDialog(null,"Failed to parse "+infoDir+entries.get(curEntry)+".xml","I/O Error",JOptionPane.ERROR_MESSAGE);
//			return false;
//		}
//		
//		
//		System.out.println("Cur entry: "+getCurEntryInImageIndex());
//		return true;
	}
	
//	private KeypointState readKeypointState(Scanner s) {
//		KeypointState state = new KeypointState();
//		state.x=s.nextDouble();
////		System.out.println("x="+state.x);
//		state.y=s.nextDouble();
////		System.out.println("y="+state.y);
//		state.z = 0;
//		if (s.hasNextDouble())
//			s.nextDouble();		// skip the Z. It will be recomputed on the fly
////		System.out.println("z="+state.z);
//		try {
//		state.zorder=ZOrder.valueOf(s.next());
//		} catch(Exception e) {
//			System.out.println("Zorder read problem:"+e);
//		}
//		String hiddenStr = s.next();
////		System.out.println("HiddenStr:"+hiddenStr);
//		assert(hiddenStr.equals(HIDDEN_STR) || hiddenStr.equals(VISIBLE_STR));
//		state.hidden = hiddenStr.equals(HIDDEN_STR);
//		return state;
//	}
	
	public boolean readCurrentAnnotation() {
		bodyType = -1;
		currentPose=null;
		currentVOCid=null;
		currentBounds=null;
		for (int i=0; i<currentLabels.length; ++i)
			currentLabels[i] = new ArrayList<Integer>();
		currentKeypoints.clear();

		InputStream instr;
		String path = new String(baseDir+infoDir+entries.get(curEntry));
		try {
			if (readOnly) {
				if (Thread.currentThread().getContextClassLoader().getResource(path)==null) {
					currentKeypoints.clear();
					return true;					
				}
				instr = Thread.currentThread().getContextClassLoader().getResourceAsStream(path);
			} else {
				File f = new File(path);
				if (!f.exists()) {
					if (!new File(baseDir+infoDir).exists()) {
						(new File(baseDir+infoDir)).mkdir();
					}
					currentKeypoints.clear();
					return true;					
				}
				instr = new FileInputStream(f);
			}
		} catch (Exception e) {
			JOptionPane.showMessageDialog(null,"Cannot read "+path,"I/O Error",JOptionPane.ERROR_MESSAGE);
			return false;
		}
		
		DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder docBuilder;
		NodeList nList;
		try {
			docBuilder = docBuilderFactory.newDocumentBuilder();
			Document doc = docBuilder.parse(instr);
			
			// ensure that the image inside matches the file name
			nList = doc.getElementsByTagName("image");
			assert(nList.getLength()==1);
			String imgName = nList.item(0).getFirstChild().getNodeValue();
			if (!imgName.equals(getCurImageName(false))) {
				JOptionPane.showMessageDialog(null,"The image inside "+path+"does not match its filename","I/O Error",JOptionPane.ERROR_MESSAGE);
				return false;				
			}

			// ensure the category matches the current one
			nList = doc.getElementsByTagName("category");
			assert(nList.getLength()==1);
			String catName = nList.item(0).getFirstChild().getNodeValue();
			if (!catName.equals(configMgr.datasetName())) {
				JOptionPane.showMessageDialog(null,"The annotation has category "+catName+" instead of "+configMgr.datasetName(),"I/O Error",JOptionPane.ERROR_MESSAGE);
				return false;				
			}
			
			nList = doc.getElementsByTagName("subcategory");
			assert(nList.getLength()<=1);
			if (nList.getLength()==1) {
				String subcatStr=nList.item(0).getFirstChild().getNodeValue();
				if (!subcatStr.equals(BODY_TYPE_UNKNOWN)) {
					for (int i=0; i<configMgr.getSubcategoryNames().length; ++i)
						if (subcatStr.equals(configMgr.getSubcategoryNames()[i]))
							bodyType=i;
					if (bodyType==-1)
						JOptionPane.showMessageDialog(null,"Cannot read subcategory from "+infoDir+entries.get(curEntry),"I/O Error",JOptionPane.ERROR_MESSAGE);					
				}
			}
			
			nList = doc.getElementsByTagName("visible_bounds");
			assert(nList.getLength()<=1);
			if (nList.getLength()==1) {
				currentBounds = new Double[4];
				currentBounds[0] = Double.valueOf(((Element)nList.item(0)).getAttribute("xmin"));
				currentBounds[1] = Double.valueOf(((Element)nList.item(0)).getAttribute("ymin"));
				currentBounds[2] = Double.valueOf(((Element)nList.item(0)).getAttribute("width"));
				currentBounds[3] = Double.valueOf(((Element)nList.item(0)).getAttribute("height"));
			}

			nList = doc.getElementsByTagName("pose");
			assert(nList.getLength()<=1);
			if (nList.getLength()==1)
				currentPose = nList.item(0).getFirstChild().getNodeValue();
			
			nList = doc.getElementsByTagName("voc_id");
			assert(nList.getLength()<=1);
			if (nList.getLength()==1)
				currentVOCid = Integer.valueOf(nList.item(0).getFirstChild().getNodeValue());

			nList = doc.getElementsByTagName("keypoints");
			assert(nList.getLength()==1);
			nList = ((Element)nList.item(0)).getElementsByTagName("keypoint");
			for (int kp=0; kp<nList.getLength(); ++kp) {
				Element elem=(Element)nList.item(kp);
				String kpName = elem.getAttribute("name");
				int keypointVal = configMgr.getKeypointID(kpName);
				if (keypointVal==-1) {
					JOptionPane.showMessageDialog(null,"Unknown keypoint "+kpName+" in "+path,"I/O Error",JOptionPane.ERROR_MESSAGE);
					return false;
				}
				KeypointState state = new KeypointState();
				state.x=Double.valueOf(elem.getAttribute("x"));
				state.y=Double.valueOf(elem.getAttribute("y"));
				state.z=Double.valueOf(elem.getAttribute("z"));
				Integer zorder = Integer.valueOf(elem.getAttribute("zorder"));
				if (zorder<0 || zorder>=ZOrder.values().length) {
					JOptionPane.showMessageDialog(null,"Unknown zorder of keypoint "+kpName+" in "+path,"I/O Error",JOptionPane.ERROR_MESSAGE);
					return false;
				}
				state.zorder=ZOrder.values()[zorder];

				int visible = Integer.valueOf(elem.getAttribute("visible")).intValue();
				if (visible!=0 && visible!=1) {
					JOptionPane.showMessageDialog(null,"Unknown visibility of keypoint "+kpName+" in "+path,"I/O Error",JOptionPane.ERROR_MESSAGE);
					return false;
				}
				state.hidden = (visible==0);			
				currentKeypoints.put(keypointVal, state);
			}

			nList = doc.getElementsByTagName("segments");
			assert(nList.getLength()<2);
			if (nList.getLength()==1) {
				nList = ((Element)nList.item(0)).getElementsByTagName("segment");
				for (int kp=0; kp<nList.getLength(); ++kp) {
					Element elem=(Element)nList.item(kp);
					String regName = elem.getAttribute("name");
					int regionID = getRegionNameID(regName);
					if (regionID<0) {
						JOptionPane.showMessageDialog(null,"Unknown region "+regName+" in "+path,"I/O Error",JOptionPane.ERROR_MESSAGE);
						return false;
					}
					String ids = elem.getAttribute("segments");
					String[] sids = ids.split("[ ]+");
					for (String s : sids)
						currentLabels[regionID].add(Integer.valueOf(s));
				}
			}
		} catch (Exception e) {
			JOptionPane.showMessageDialog(null,"Cannot read "+path,"I/O Error",JOptionPane.ERROR_MESSAGE);
			return false;
		}

		return true;
	}

	public boolean writeCurrentAnnotation() {
		if (readOnly) {
			JOptionPane.showMessageDialog(null,"The demo files are read-only","I/O Error",JOptionPane.ERROR_MESSAGE);
			return false;
		}
        DocumentBuilderFactory dbfac = DocumentBuilderFactory.newInstance();
        DocumentBuilder docBuilder;
		try {
			java.text.DecimalFormat format = new java.text.DecimalFormat("0.00",new DecimalFormatSymbols(Locale.US));

			docBuilder = dbfac.newDocumentBuilder();
	        Document doc = docBuilder.newDocument();
	        
			Element root = doc.createElement("annotation");
			doc.appendChild(root);
			
			Element imgNode = doc.createElement("image");
			imgNode.appendChild(doc.createTextNode(getCurImageName(false)));
			root.appendChild(imgNode);

			Element catNode = doc.createElement("category");
			catNode.appendChild(doc.createTextNode(configMgr.datasetName()));
			root.appendChild(catNode);

			if (bodyType>-1) {
				Element subcatNode = doc.createElement("subcategory");
				subcatNode.appendChild(doc.createTextNode(configMgr.getSubcategoryNames()[bodyType]));
				root.appendChild(subcatNode);
			}
			if (currentPose!=null) {
				Element poseNode = doc.createElement("pose");
				poseNode.appendChild(doc.createTextNode(currentPose));
				root.appendChild(poseNode);				
			}
			if (currentBounds!=null) {
				Element boundsNode = doc.createElement("visible_bounds");
				boundsNode.setAttribute("xmin", format.format(currentBounds[0]));
				boundsNode.setAttribute("ymin", format.format(currentBounds[1]));
				boundsNode.setAttribute("width", format.format(currentBounds[2]));
				boundsNode.setAttribute("height", format.format(currentBounds[3]));
				root.appendChild(boundsNode);
			}
			if (currentVOCid!=null) {
				Element vocNode = doc.createElement("voc_id");
				vocNode.appendChild(doc.createTextNode(currentVOCid.toString()));
				root.appendChild(vocNode);				
			}
			Element kpsNode = doc.createElement("keypoints");
			Map<String, KeypointState> kpsSortedByName = new TreeMap<String,KeypointState>();
			for (Map.Entry<Integer, KeypointState> entry : currentKeypoints.entrySet())
				if (!configMgr.getKeypointIsHelper()[entry.getKey()])
					kpsSortedByName.put(configMgr.getKeypointNames()[entry.getKey()], entry.getValue());

			for (Map.Entry<String, KeypointState> entry : kpsSortedByName.entrySet()) {
				Element kpNode = doc.createElement("keypoint");
				kpNode.setAttribute("name", entry.getKey());
				kpNode.setAttribute("visible", entry.getValue().hidden ? "0":"1");
				kpNode.setAttribute("x",format.format(entry.getValue().x));
				kpNode.setAttribute("y",format.format(entry.getValue().y));
				kpNode.setAttribute("z",format.format(entry.getValue().z));
				kpNode.setAttribute("zorder",String.valueOf(entry.getValue().zorder.ordinal()));
				kpsNode.appendChild(kpNode);
			}
			root.appendChild(kpsNode);
			
			Map<String, ArrayList<Integer>> regsSortedByName = new TreeMap<String,ArrayList<Integer>>();
			for (int label=0; label<currentLabels.length; ++label) {
				if (!currentLabels[label].isEmpty())
				regsSortedByName.put(regionNames[label],currentLabels[label]);
			}
			if (!regsSortedByName.isEmpty()) {
				Element segsNode = doc.createElement("segments");
				for (Map.Entry<String, ArrayList<Integer>> entry : regsSortedByName.entrySet()) {
					Element segNode = doc.createElement("segment");
					segNode.setAttribute("name",entry.getKey());
					String ids=new String();
					for (Integer elem : entry.getValue())
						ids+=elem+" ";
					segNode.setAttribute("segments", ids.substring(0,ids.length()-1));
					segsNode.appendChild(segNode);
				}
				root.appendChild(segsNode);
			}

	        // write the XML document to disk
	        Source xmlSource = new DOMSource(doc);
	        Result result = new StreamResult(new FileOutputStream((baseDir+infoDir+entries.get(curEntry))));
	        TransformerFactory transformerFactory = TransformerFactory.newInstance();
	        Transformer transformer = transformerFactory.newTransformer();
	        transformer.setOutputProperty(OutputKeys.INDENT, "yes");
	        transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
	        transformer.transform(xmlSource, result);
        } catch (IOException e) {
			JOptionPane.showMessageDialog(null,"Cannot write "+infoDir+entries.get(curEntry)+".xml","I/O Error",JOptionPane.ERROR_MESSAGE);
			return false;
		} catch (ParserConfigurationException e) {
			JOptionPane.showMessageDialog(null,"Cannot write "+infoDir+entries.get(curEntry)+".xml","I/O Error",JOptionPane.ERROR_MESSAGE);
			return false;
		} catch (TransformerException e) {
			JOptionPane.showMessageDialog(null,"Cannot write "+infoDir+entries.get(curEntry)+".xml","I/O Error",JOptionPane.ERROR_MESSAGE);
			return false;
		}

        return true;
	}
	public boolean writeKeypoints() {
		return writeCurrentAnnotation();
//		if (readOnly) {
//			JOptionPane.showMessageDialog(null,"The demo files are read-only","I/O Error",JOptionPane.ERROR_MESSAGE);
//			return false;
//		}
//		
//		PrintWriter writer;
//		try {
//			writer = new PrintWriter(new FileWriter((baseDir+infoDir+entries.get(curEntry)+".xml")));
//			java.text.DecimalFormat format = new java.text.DecimalFormat("0.00",new DecimalFormatSymbols(Locale.US));
//			String[] writeLines = new String[currentKeypoints.size()];
//			int i=0;
//			for (Map.Entry<Integer, KeypointState> entry : currentKeypoints.entrySet()) {
//				KeypointState ks=entry.getValue();
//				writeLines[i++] = configMgr.getKeypointNames()[entry.getKey()]+" "+format.format(ks.x)+" "+format.format(ks.y)+" "+format.format(ks.z)+" "+ks.zorder+" "+(ks.hidden ? HIDDEN_STR : VISIBLE_STR)+"\n";
//			}
//			java.util.Arrays.sort(writeLines);
//			
//			for (String str : writeLines)
//				writer.write(str);
//			
//			writer.write(BODY_TYPE+" "+(bodyType==-1 ? BODY_TYPE_UNKNOWN : configMgr.getSubcategoryNames()[bodyType])+"\n");
//
//			writeLines = new String[currentLabels.length];
//			for (int label=0; label<currentLabels.length; ++label) {
//				ArrayList<Integer> sp = currentLabels[label];
//				if (sp.isEmpty()) {
//					writeLines[label]=new String();
//					continue;
//				}
//				String str=new String(regionNames[label]+" ");
//				
////				writer.write(regionNames[label]+" ");
//				for (int id : sp)
//					str+=(id + " ");
////					writer.write(id + " ");
//				str+="\n";
////				writer.write("\n");
//				writeLines[label]=str;
//			}
//			java.util.Arrays.sort(writeLines);
//			
//			for (String str : writeLines)
//				if (str.length()>0)
//					writer.write(str);
//				
//		    writer.close();
//		} catch (IOException e) {
//			JOptionPane.showMessageDialog(null,"Cannot write "+infoDir+entries.get(curEntry)+".xml","I/O Error",JOptionPane.ERROR_MESSAGE);
//			return false;
//		}
//		return true;
	}
	
	private boolean readAllEntries() {
		entries = new ArrayList<String>();
		
		BufferedReader reader;
		try {
			if (readOnly) {
				reader = new BufferedReader(new InputStreamReader(Thread.currentThread().getContextClassLoader().getResourceAsStream(baseDir+allEntriesFileName)));
				String line=null;
				while ((line=reader.readLine())!=null) {
					entries.add(line);
				}
				reader.close();				
			} else {
				File allEntriesFile = new File(baseDir+allEntriesFileName);
				if (allEntriesFile.exists()) {
					reader = new BufferedReader(new FileReader(allEntriesFile));			
					String line=null;
					while ((line=reader.readLine())!=null) {
						entries.add(line);
					}
					reader.close();				
				} else {
					// the index file does not exist - try to recreate it. First try using info directory.
					File infoDirName = new File(baseDir+infoDir);
					if (infoDirName.exists()) {
						File[] strFilesDirs = infoDirName.listFiles();
						for (File f : strFilesDirs) {
							String fileName = f.getName().toLowerCase();
							if (f.getName().endsWith(".xml"))
								entries.add(f.getName());
						}		
					} else {
						// The index files doesn't exist and there are no annotations. Create a new list from the images.
						File imagesDir = new File(baseDir+imgDir);
						if (!imagesDir.exists())
							throw new Exception("No image directory");
						File[] strFilesDirs = imagesDir.listFiles();
						for (File f : strFilesDirs) {
							String fileName = f.getName().toLowerCase();
							if (fileName.endsWith(".jpg"))
								entries.add(fileName.substring(0,fileName.length()-4)+"_0.xml");
						}						
					}
					writeAllEntries();					
				}
			}
		} catch (Exception e) {
			File imagesDir = new File(baseDir+imgDir);
			if (!imagesDir.exists()) {
				JOptionPane.showMessageDialog(null,"Cannot read "+baseDir+allEntriesFileName,"I/O Error",JOptionPane.ERROR_MESSAGE);
				return false;
			}
		}
		return true;
	}
	
	private boolean writeAllEntries() {
		File allFiles = new File(baseDir+allEntriesFileName);

		BufferedWriter writer;
		try {
			writer = new BufferedWriter(new FileWriter(allFiles));
			
			for (String line : entries) {
				writer.write(line+"\n");
			}
			writer.close();

		} catch (Exception e) {
			JOptionPane.showMessageDialog(null,"Cannot write "+allEntriesFileName,"I/O Error",JOptionPane.ERROR_MESSAGE);
			return false;
		}
		return true;
	}
	
	public int getCurEntryInImageIndex() {
		String str = entries.get(curEntry);
		int sep = str.lastIndexOf('_');
		if (sep<0) {
			JOptionPane.showMessageDialog(null,"Entry "+str+" has invalid format.","I/O Error",JOptionPane.ERROR_MESSAGE);
			return 0;
		}
		try { 
			return Integer.parseInt(str.substring(sep+1,(str.length()-4)));
		} catch (NumberFormatException e) {
			JOptionPane.showMessageDialog(null,"Entry "+str+" has invalid format.","I/O Error",JOptionPane.ERROR_MESSAGE);
			return 0;
		}
	}
	
	ArrayList<Integer>[] getRegionLabels() {
		return currentLabels;
	}
	
	void setRegionLabels(ArrayList<Integer>[] labels) {
		currentLabels = labels;
	}
	

	public String getCurImageName(boolean withextension) {
		return getEntryImageName(entries.get(curEntry),withextension);
	}
	private String getEntryImageName(String entry,boolean withextension) {
		int sep = entry.lastIndexOf('_');
		if (sep<0) {
			JOptionPane.showMessageDialog(null,"Entry "+entry+" has invalid format.","I/O Error",JOptionPane.ERROR_MESSAGE);
			return "";
		}
		String ret = entry.substring(0,sep);
		return withextension ? ret+".jpg" : ret;
	}
	
	// returns whether there exists at least one entry that refers to the given image
	private boolean entriesContainImageName(String imgName) {
		for (String str : entries) {
			if (getEntryImageName(str,false).equals(imgName))
				return true;
		}
		return false;
	}
	
	private int getRegionNameID(String name) {
		for (int i=0; i<regionNames.length; ++i)
			if (regionNames[i].equals(name))
				return i;
		return -1;
	}

	public IOMgr(String baseDir, String imgDir, String infoDir, String pbDir, String allEntriesFileName, boolean readOnly, ConfigMgr configMgr) {
		this.baseDir = baseDir;
		this.infoDir = infoDir;
		this.imgDir = imgDir;
		this.pbDir = pbDir;
		this.allEntriesFileName = allEntriesFileName;
		this.readOnly = readOnly;
		this.configMgr = configMgr;
		regionNames = configMgr.getRegionNames();
		currentLabels = new ArrayList[regionNames.length];

		readAllEntries();
		curEntry=0;		
		syncFromDisk();
	}
}
